#include "defs.h"
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

int getrand(int max)
{
  int i = (double)rand()/RAND_MAX * (max+1);
  if (i == max+1) i--;
  return i;
}

void print_state(state *s)
{
  int i, j, n;
  int m[s->n][s->n];
  for (n = 0; n < s->n * s->n; n++) {
    m[s->row[n]][s->col[n]] = n+1;
  }
  for (i = 0; i < s->n; i++) {
    printf("(");
    for (j = 0; j < s->n; j++) {
      printf("%d", m[i][j]);
      if (j != s->n-1) printf(",");
    }
    printf(")");
    if (i != s->n-1) printf(",");
    printf("\n");
  }
}

void print_best(problem *p)
{
  int i;
  printf("best e:\n");
  for (i = 3; i <= 27; i++)
    printf("%d: min %ld max %ld\n", i, p->s[i*2].best.e, p->s[i*2+1].best.e);
  for (i = 3; i <= 27; i++) {
    print_state(&p->s[i*2].best);
    printf(";\n");
    print_state(&p->s[i*2+1].best);
    if (i != 27) printf(";\n");
  }
}

void save_state(FILE *f, state *s)
{
  int i, j, n;
  int m[s->n][s->n];
  for (n = 0; n < s->n * s->n; n++) {
    m[s->row[n]][s->col[n]] = n+1;
  }
  for (i = 0; i < s->n; i++) {
    fprintf(f, "(");
    for (j = 0; j < s->n; j++) {
      fprintf(f, "%d", m[i][j]);
      if (j != s->n-1) fprintf(f, ",");
    }
    fprintf(f, ")");
    if (i != s->n-1) fprintf(f, ",");
    fprintf(f, "\n");
  }
}

void save_best(problem *p, char *name)
{
  FILE *f = fopen(name, "w"); if (f == NULL) abort();
  int i;
  fprintf(f, "best e:\n");
  for (i = 3; i <= 27; i++)
    fprintf(f, "%d: min %ld max %ld\n",
            i, p->s[i*2].best.e, p->s[i*2+1].best.e);
  for (i = 3; i <= 27; i++) {
    save_state(f, &p->s[i*2].best);
    fprintf(f, ";\n");
    save_state(f, &p->s[i*2+1].best);
    if (i != 27) fprintf(f, ";\n");
  }
  fclose(f);
}
extern int dist[28][28][28][28];
extern int gcd[27*27+1][27*27+1];

void e(state *s)
{
  int n0, n1, i0, j0, i1, j1;
  long g, d, e = 0;
  for (n0 = 0; n0 < s->n * s->n; n0++) {
    i0 = s->row[n0];
    j0 = s->col[n0];
    for (n1 = n0 + 1; n1 < s->n * s->n; n1++) {
      i1 = s->row[n1];
      j1 = s->col[n1];
      g = gcd[n0+1][n1+1];
      d = dist[i0][j0][i1][j1];
      e += g * d;
    }
  }
  s->e = e;
}

void init_sub(sub *s, int n, enum type t)
{
  int nn;
  s->cur.t = s->best.t = t;
  s->cur.n = s->best.n = n;
  for (nn = 0; nn < n * n; nn++) {
    s->best.row[nn] = s->cur.row[nn] = nn / n;
    s->best.col[nn] = s->cur.col[nn] = nn % n;
  }
  e(&s->cur);
  e(&s->best);
  s->s.k = 0;
  s->s.kmax = n * 10000;
}

void init_problem(problem *p)
{
  int si;

  for (si = 3; si <= 27; si++) {
    init_sub(&p->s[si*2], si, TMIN);
    init_sub(&p->s[si*2+1], si, TMAX);
  }
}

void init(void);

void neighbor(state *n, state *s)
{
  int n0 = getrand(s->n * s->n - 1);
  int n1;
  do n1 = getrand(s->n * s->n - 1); while (n1 == n0);
  *n = *s;
  /* exchange elements */
  n->row[n0] = s->row[n1];
  n->col[n0] = s->col[n1];
  n->row[n1] = s->row[n0];
  n->col[n1] = s->col[n0];
  e(n);
}

int _sup(long a, long b)
{
  return a > b;
}

int _inf(long a, long b)
{
  return a < b;
}

int newbest;

int do_sa(sub *s)
{
  double temp;
  state new;
  double prand, p;
  int (*inf)(long, long);

  if (s->s.k == s->s.kmax) return 0;

  temp = (double)s->s.k / s->s.kmax;
  temp = (1 - temp) * 100;

  neighbor(&new, &s->cur);

  if (s->cur.t == TMIN) inf = _inf; else inf = _sup;

  prand = (double)rand()/((double)(RAND_MAX)-1); if (prand > 1) prand = 1;

  if (inf(new.e, s->cur.e)) p = 1;
  else p = exp(-abs(new.e - s->cur.e) / temp);

  if (p == 1 || p > prand) {
    s->cur = new;
    if (inf(new.e, s->best.e)) { s->best = new; newbest = 1; }
  }

  s->s.k++;

  return 1;
}

void usage(void)
{
  printf("usage: <options> [best file 1] .. [best file n]\n");
  printf("options:\n");
  printf("    -post\n");
  printf("        only do a post processing\n");
  printf("        get min/max at each turn up to no more min/max\n");
  exit(1);
}

problem p;
int main(int n, char **v)
{
  char *best[n];
  int nbest = 0;
  int do_post = 0;
  int i;
  int somedone;
  init();
  init_problem(&p);

  /* you can pass "best" files to update best */
  for (i = 1; i < n; i++) {
    if (!strcmp(v[i], "-post")) { do_post = 1; continue; }
    if (v[i][0] == '-') usage();
    best[nbest] = v[i];
    nbest++;
  }

  for (i = 0; i < nbest; i++) update_best(&p, best[i]);

  if (do_post) { post(&p); return 0; }

  do {
    somedone = 0;
    for (i = 3; i <= 27; i++) {
      somedone += do_sa(&p.s[i*2]);
      somedone += do_sa(&p.s[i*2+1]);
    }
    if (newbest) { save_best(&p, "best.txt"); newbest = 0; }
  } while (somedone);
  print_best(&p);
  return 0;
}
